;===============================================================================
;	GIFLIB v1.0 - (C) Copyright by Exagone
;===============================================================================
; The included general public license (see license.txt) applies to this source
; code. By using this code, you agree to the terms of the license.
; 
; Original source code by Exagone
; http://exagone.cjb.net | http://www.exagone.org
; _thomas_@mailroom.com
;
; Libraries and include files from the Win32Asm package by Iczelion:
; http://win32asm.cjb.net/
.486
.model flat, stdcall
option casemap:none

includelib 	kernel32.lib
includelib  user32.lib
includelib 	gdi32.lib

include		\masm32\include\kernel32.inc
include		\masm32\include\user32.inc
include		\masm32\include\gdi32.inc
include		\masm32\include\windows.inc
include		giflib_internal.inc

GIFLoadHeader			PROTO	STDCALL :DWORD
GIFInitalizeDecoder		PROTO	STDCALL :DWORD
GIFDecompress			PROTO	STDCALL :DWORD
SkipExtension			PROTO	STDCALL :DWORD, :DWORD
SkipSubBlocks			PROTO	STDCALL :DWORD
ProcessImageDescriptor	PROTO	STDCALL :DWORD, :DWORD, :DWORD
DECInitialize			PROTO	STDCALL :DWORD
DECResetLZW				PROTO	STDCALL :DWORD
DECDecodeGIF			PROTO	STDCALL :DWORD
DECIsCodeInTable		PROTO	STDCALL :DWORD, :DWORD
DECOutputString			PROTO	STDCALL :DWORD, :DWORD
GetNextCode				PROTO	STDCALL :DWORD
LoadNextByte			PROTO	STDCALL :DWORD
DECAddToTable			PROTO	STDCALL :DWORD, :DWORD, :DWORD
GIFGetMinimumColorType	PROTO	STDCALL :DWORD
GIFGetOutputSize		PROTO	STDCALL :DWORD
GIFGetColorTable		PROTO	STDCALL :DWORD, :DWORD
GIFCleanup				PROTO	STDCALL :DWORD
GIFFillBitmapInfoStruct	PROTO	STDCALL :DWORD, :DWORD
GIFLoadResource			PROTO	STDCALL :DWORD, :DWORD, :DWORD
GIFLoadFile				PROTO	STDCALL :DWORD, :DWORD
ClearMemory				PROTO	STDCALL :DWORD, :DWORD

.code
;===============================================================================
;	ClearMemory	
;===============================================================================
; Zeroes memory
ClearMemory		proc	lpMem:DWORD, dwSize:DWORD
    cld
    mov edi, lpMem
    mov ecx, dwSize
	xor	eax, eax
    shr ecx, 2
    rep stosd

    mov ecx, dwSize
    and ecx, 3
    xor	al, al
    rep stosb
    ret
ClearMemory		endp
;===============================================================================
;	GIFLoadResource
;===============================================================================
; Loads gif data from a resource (of type RT_RCDATA)
; lpName:pointer to a resource name OR resource ID
GIFLoadResource proc	lpGifInfo:DWORD, hInst:DWORD, lpName:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	
	invoke	FindResource, hInst, lpName, RT_RCDATA
	.IF		eax==0
		ret
	.ENDIF
	push	eax
	invoke	LoadResource, hInst, eax
	.IF		eax==0
		ret
	.ENDIF
	mov		[esi].lpGIFData, eax
	pop		eax
	invoke	SizeofResource, hInst, eax
	.IF		eax==0
		ret
	.ENDIF
	mov		[esi].dwGIFDataSize, eax
	assume	esi:nothing
xor		eax, eax
inc		eax
ret
GIFLoadResource	endp

;===============================================================================
;	GIFLoadFile
;===============================================================================
; Loads gif data from a file
; SHOULD CLEANUP POINTER AFTERWARDS (globalalloc)
GIFLoadFile		proc lpGifInfo:DWORD, lpFile:DWORD
LOCAL	hFile:DWORD
LOCAL	pMem:DWORD
LOCAL	bRead:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	invoke	CreateFile, lpFile, GENERIC_READ, FILE_SHARE_READ, NULL,\
				OPEN_EXISTING, NULL, NULL
	.IF		eax==INVALID_HANDLE_VALUE
		xor		eax, eax
		ret
	.ENDIF
	mov		hFile, eax
	invoke	GetFileSize, eax, NULL
	.IF		eax==-1
		invoke	CloseHandle, hFile
		xor		eax, eax
		ret
	.ENDIF
	mov		[esi].dwGIFDataSize, eax
	invoke	GlobalAlloc, GMEM_FIXED, eax
	.IF		eax==0
		invoke	CloseHandle, hFile
		xor		eax, eax
		ret
	.ENDIF
	mov		pMem, eax
	invoke	ReadFile, hFile, pMem, [esi].dwGIFDataSize, ADDR bRead, NULL
	.IF		eax==NULL
		invoke	CloseHandle, hFile
		xor		eax, eax
		ret
	.ENDIF
	push	pMem
	pop		[esi].lpGIFData
	invoke	CloseHandle, hFile
	xor		eax, eax
	inc		eax
	assume	esi:nothing
ret
GIFLoadFile		endp
;===============================================================================
;	GIFCleanup
;===============================================================================
; Cleanup
; 
GIFCleanup		proc	uses esi lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	invoke	GlobalFree, [esi].lpLZWTable
	assume	esi:nothing
ret
GIFCleanup		endp

;===============================================================================
;	GIFGetColorTable
;===============================================================================
; Retrieves the color table to a given buffer, in the RGBQUAD format
; 
GIFGetColorTable	proc	uses esi edi ebx lpGifInfo:DWORD, lpBuffer:DWORD
	
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	
	mov		edi, lpBuffer
	mov		ebx, [esi].lpColorTable
	xor		ecx, ecx
	.WHILE	ecx<[esi].ColorCount
		mov		eax, [ebx]	; RR GG BB 00 -> eax: 00 BB GG RR
		bswap 	eax			; eax: RR GG BB 00
		shr		eax, 8		; eax: 00 RR GG BB (RGBQUAD)
		stosd 
		add		ebx, 3
	inc	ecx
	.ENDW
	assume	esi:nothing
ret
GIFGetColorTable	endp
;===============================================================================
;	GIFGetOutputSize
;===============================================================================
; Gets the number of bytes to reserve for the output data
GIFGetOutputSize	proc	uses esi ebx 	lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	
	mov		eax, [esi].dwImageWidth
	mov		ecx, eax

	; ---Get aligned width ---
	.IF [esi].dwImageFormat==IMAGEFORMAT_1BIT	
		and		ecx, 11111b
		.IF		ecx!=0		;
			shr		eax,5	
			inc		eax		
			shl		eax, 5	;get nr of bytes
		.ENDIF
	.ELSEIF	[esi].dwImageFormat==IMAGEFORMAT_4BIT	
		
		and		ecx, 111b
		.IF		ecx!=0		;
			shr		eax,3	
			inc		eax		
			shl		eax,3	;get nr of bytes
		.ENDIF
	.ELSEIF [esi].dwImageFormat==IMAGEFORMAT_8BIT
		and		ecx, 11b
		.IF		ecx!=0
			shr		eax, 2
			inc		eax
			shl		eax, 2
		.ENDIF
	.ENDIF 
	
	; --- multiply output width by nr of scanlines ---
	mul		[esi].dwImageHeight
	
	assume	esi:nothing
ret
GIFGetOutputSize	endp

;===============================================================================
;	GIFFillBitmapInfoStruct
;===============================================================================
;  Fills a bitmap info structure width the given properties or automatically
GIFFillBitmapInfoStruct	proc	uses esi edi	lpGifInfo:DWORD, lpBIH:DWORD
	mov		esi, lpGifInfo
	mov		edi, lpBIH
	assume	esi:PTR GIF_INFO
	assume	edi:PTR BITMAPINFOHEADER
	mov		[edi].biSize, SIZEOF BITMAPINFOHEADER
	mov		eax,[esi].dwImageHeight
	neg		eax	;negative because it outputs a top-down bitmap
	mov		[edi].biHeight, eax
	push	[esi].dwImageWidth
	pop		[edi].biWidth
	
	mov		[edi].biPlanes, 1
	mov		eax, [esi].dwImageFormat
	.IF		eax==IMAGEFORMAT_1BIT
		mov		ax, 1
	.ELSEIF eax==IMAGEFORMAT_4BIT
		mov		ax, 4
	.ELSE
		mov		ax, 8
	.ENDIF
	mov		[edi].biBitCount, ax
	xor		eax, eax
	mov		[edi].biCompression, BI_RGB	
	mov		[edi].biSizeImage, eax
	mov		[edi].biXPelsPerMeter, eax
	mov		[edi].biYPelsPerMeter, eax
	push	[esi].ColorCount
	pop		[edi].biClrUsed
	mov		[edi].biClrImportant, eax

	assume	esi:nothing
	assume	edi:nothing
ret
GIFFillBitmapInfoStruct	endp
;===============================================================================
;	GIFGetMinimumColorType
;===============================================================================
; Gets the minimum color type (IMAGEFORMAT_XXX) to use for outputting the
; image.
GIFGetMinimumColorType	proc	uses esi	lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	mov		eax, [esi].ColorCount
	.IF		eax==2
		mov		eax, IMAGEFORMAT_1BIT
	.ELSEIF	eax<=16
		mov		eax, IMAGEFORMAT_4BIT
	.ELSEIF eax<=256
		mov		eax, IMAGEFORMAT_8BIT
	.ENDIF
	assume	esi:nothing
ret
GIFGetMinimumColorType	endp
;===============================================================================
;	GIFDecompress
;===============================================================================
; Decompresses the gif and outputs
; returns:
; 0  - color format not allowed
; 1  - success
; -1 - failed
GIFDecompress	proc	uses esi edi ebx lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	mov		edi, [esi].lpImageData
	; --- Check if color format is allowed ---
	mov		eax, [esi].dwImageFormat
	.IF		eax<[esi].ColorCount
		xor		eax, eax
		ret
	.ENDIF	
	
	; --- Initially, load 4 bytes ---
	mov		[esi].CurrentBits, 0
	mov		[esi].dbBitsAvailable, 0
	mov		ebx,4
	.WHILE	ebx>0
		invoke	LoadNextByte, esi
		dec 	ebx
	.ENDW
	
	invoke	DECInitialize, esi
	invoke	DECDecodeGIF, lpGifInfo
	
	assume	esi:nothing
xor	eax, eax
inc eax
ret
GIFDecompress	endp
;===============================================================================
;	DECDecodeGIF
;===============================================================================
; The actual decoder
;
DECDecodeGIF	proc	uses esi edi ebx lpGifInfo:DWORD
LOCAL	OldCode:DWORD
LOCAL	FirstCharOfOldCode:BYTE
LOCAL	FirstCharOfCode:BYTE
LOCAL	NrCodesToOutput:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	
	; --- set counters ---
	mov		[esi].PixelCounter, 0	;0 codes outputted
	mov		eax, [esi].dwImageWidth
	mul		[esi].dwImageHeight
	mov		NrCodesToOutput, eax	;this is an OUTPUT CODE counter, not nr of compressed codes
	; --- clear output mem ---
	invoke	GIFGetOutputSize, lpGifInfo
	mov		ecx, eax
	invoke	ClearMemory, [esi].lpImageData, ecx
	
	
;	 LZW Decompression
;     [1] Initialize string table;
;     [2] get first code: <code>;
;     [3] output the string for <code> to the charstream;
;     [4] <old> = <code>;
;     [5] <code> <- next code in codestream;
;     [6] does <code> exist in the string table?
;      (yes: output the string for <code> to the charstream; (A)
;            [...] <- translation for <old>;				 (B)
;            K <- first character of translation for <code>; (C)
;            add [...]K to the string table;    			 (D)
;			 <old> <- <code>;  )							 (E)
;      (no: [...] <- translation for <old>;					 (F)
;           K <- first character of [...];					 (G)
;           output [...]K to charstream and add it to string table; (H)
;           <old> <- <code>									 (I)
;      )
;     [7] go to [5];

	.WHILE	TRUE							
		mov		eax, NrCodesToOutput
		.IF		[esi].PixelCounter==eax
			.BREAK	;done
		.ENDIF
		invoke	GetNextCode, esi			; [1] and [5]
		mov		ebx, eax
		.IF		ebx==[esi].dwLZWClear		; Check for clearcode
			invoke	DECResetLZW, esi		; Reset LZW
			.CONTINUE
		.ENDIF
		.IF		[esi].IsFirstCode==1		; Checks if steps [1]-[4] have to be done
			dec		[esi].IsFirstCode		  ;[2]
			invoke	DECOutputString, esi, ebx ;[3]
			mov		FirstCharOfOldCode, al	  ; saved for (G)
			mov		OldCode, ebx			  ;[4]
			.CONTINUE						  ;[5]
		.ENDIF
		invoke	DECIsCodeInTable, esi, ebx	  ;[6]
		.IF		eax==TRUE					  ;(A)-(E)
			invoke	DECOutputString, esi, ebx 	;(B)
			mov		FirstCharOfCode, al			;(C)
			mov		FirstCharOfOldCode, al		; saved for (G)
			invoke	DECAddToTable, esi, OldCode, eax ;(D)
			mov		OldCode, ebx
		.ELSE								  ;(F)-(I)
			mov		al, FirstCharOfOldCode		; (G)
			invoke	DECAddToTable, esi, OldCode, eax ;(H)
			invoke	DECOutputString, esi, ebx   ;(H)
			mov		FirstCharOfOldCode, al		; saved for (G)
			mov		OldCode, ebx
		.ENDIF
	.ENDW									  ;[7]
	
	assume	esi:nothing
xor	eax, eax
inc eax
ret
DECDecodeGIF	endp


;===============================================================================
; DECAddToTable
;===============================================================================
; Adds an entry to the LZW Table
; The enry consists of a prefix and a suffix (size is calculated automatically)
; returns the code for the new entry
DECAddToTable proc uses esi edi ebx lpGifInfo:DWORD, dwPrefix:DWORD, dbSuffix:DWORD
LOCAL	NewMem:DWORD
LOCAL	OldSize:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	
	
	; --- Check if enough space in the table left
	mov		ecx, [esi].dwLZWTableSize
	mov		eax, [esi].dwLZWTableCount
	inc		eax
	shl		eax, 2
	; --- If no space left, make the table bigger:
	.IF		eax==ecx
		mov		OldSize, ecx
		shl		ecx, 1		;Table size doubles
		mov		[esi].dwLZWTableSize, ecx
		; --- Reallocate memory ---
		mov		eax, [esi].lpLZWTable
		invoke	GlobalAlloc, GMEM_FIXED, ecx
		mov		NewMem, eax
		; --- Copy old contents to new mem ---
		push	esi
		push	edi
		mov		esi, [esi].lpLZWTable
		mov		edi, NewMem
		mov		ecx, OldSize
		shr		ecx, 2
		rep		movsd
		pop		edi
		pop		esi
		invoke	GlobalFree, [esi].lpLZWTable
		push	NewMem
		pop		[esi].lpLZWTable
	.ENDIF
	
	; --- Create pointer to first free entry ---
	mov		ebx, [esi].dwLZWTableCount
	push	ebx
	shl		ebx, 2
	add		ebx, [esi].lpLZWTable
	
	; --- Find out the size of the string ---
	mov		eax, dwPrefix
	shl		eax, 2
	add		eax, [esi].lpLZWTable
	mov		eax, [eax]
	shr 	eax, 8
	and		eax, 0FFFh	;eax now holds the size of the prefix string
	inc		eax
	
	; --- Create the entry DWORD ---
	; (see DECResetLZW for the format)
	mov		ecx, dwPrefix
	shl		ecx, 20
	shl		eax, 8
	or		ecx, eax
	mov		eax, dbSuffix
	and		eax, 0ffh
	or		ecx, eax
	
	; --- Store entry ---
	mov		[ebx], ecx

	inc		[esi].dwLZWTableCount
	pop		ebx	;get saved code of new entry
	mov		cl, [esi].dbCurBitSize
	xor		eax, eax
	inc		eax
	shl		eax, cl
	dec		eax
	.IF		ebx==eax	;bitsize should be increased
		inc		[esi].dbCurBitSize	;increase bitsize
	.ENDIF
	
	mov		eax, ebx
ret
DECAddToTable endp


;===============================================================================
; DECOutputString
;===============================================================================
; Writes the string of dwCode to the output stream
; returns the first character of the string in al
DECOutputString		proc uses esi edi  ebx lpGifInfo:DWORD, dwCode:DWORD
LOCAL	LastChar:BYTE
LOCAL	AlignedWidth:DWORD		;image width DWORD  ALIGNED (scanline aligned width)
LOCAL	ImageWidth:DWORD		;actual image width (scanline width)
LOCAL	BitOffset:BYTE
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO

	
	; --- Set widths ---
	mov		eax, [esi].dwImageWidth
	mov		ImageWidth, eax
	mov		ecx, eax
	
	; --- Set aligned width
	.IF [esi].dwImageFormat==IMAGEFORMAT_1BIT	
		and		ecx, 11111b
		.IF		ecx!=0		;
			shr		eax,5	
			inc		eax		
			shl		eax, 5	;get nr of bytes
		.ENDIF
	.ELSEIF	[esi].dwImageFormat==IMAGEFORMAT_4BIT	
		
		and		ecx, 111b
		.IF		ecx!=0		;
			shr		eax,3	
			inc		eax		
			shl		eax,3	;get nr of bytes
		.ENDIF
	.ELSEIF [esi].dwImageFormat==IMAGEFORMAT_8BIT
		and		ecx, 11b
		.IF		ecx!=0
			shr		eax, 2
			inc		eax
			shl		eax, 2
		.ENDIF
	.ENDIF 
	mov		AlignedWidth, eax
	
	; --- Set basepointer ---
	mov		edi, [esi].lpLZWTable
	
	; --- Get entry ---
	mov		eax, dwCode
	mov		ebx, [edi+4*eax]
	
	; --- Get size of string ---
	shr		ebx, 8
	and		ebx, 0fffh
	; --- 8-BIT output ---
	.IF		[esi].dwImageFormat==IMAGEFORMAT_8BIT
		; --- 1-BIT output ---
		mov		edx, [esi].lpImageData
		push	ebx

		; --- Output all codes backwards ---
		mov		eax, dwCode
		dec		ebx				;use ebx as counter
		
		.WHILE	TRUE
			mov		eax, [edi+4*eax]
			mov		LastChar, al		
			push	eax
			push	edx
			;--- PixelOffset = (PixelCounter/width)*alignedwidth + (PixelCounter % width) ---
			mov		eax, ebx
			add		eax, [esi].PixelCounter

			cdq
			div		ImageWidth
			mov		ecx, edx
			mul		AlignedWidth
			add		ecx, eax
			
			mov		al, LastChar
			;RealByteOffset = PixelOffset 
			;RealBitOffset = 0
			pop		edx
			or		[edx+ecx], al
			
			
		
			pop		eax
			shr		eax, 20
			and		eax, 0fffh
			.BREAK .IF ebx==0
			dec		ebx
		.ENDW
		pop		ebx
		add		[esi].PixelCounter, ebx
		mov		al, LastChar

	.ELSEIF [esi].dwImageFormat==IMAGEFORMAT_4BIT	
	; --- 4-BIT output ---
		mov		edx, [esi].lpImageData
		push	ebx

		; --- Output all codes backwards ---
		mov		eax, dwCode
		dec		ebx				;use ebx as counter
		
		.WHILE	TRUE
			mov		eax, [edi+4*eax]
			mov		LastChar, al		
			push	eax
			push	edx
			;--- PixelOffset = (PixelCounter/width)*alignedwidth + (PixelCounter % width) ---
			mov		eax, ebx
			add		eax, [esi].PixelCounter

			cdq
			div		ImageWidth
			mov		ecx, edx
			mul		AlignedWidth
			add		ecx, eax
			mov		edx, ecx
			
			pop     edx
			;====
			mov		al, LastChar
			shr		ecx, 1
			
			.IF		!CARRY?		;use high bits
				shl		al, 4
				or		[edx+ecx], al
			.ELSE
				or		[edx+ecx], al			
			.ENDIF
			
			pop		eax
			shr		eax, 20
			and		eax, 0fffh
			.BREAK .IF ebx==0
			dec		ebx
		.ENDW
		pop		ebx
		add		[esi].PixelCounter, ebx
		mov		al, LastChar
	.ELSEIF [esi].dwImageFormat==IMAGEFORMAT_1BIT	
	; --- 1-BIT output ---
		mov		edx, [esi].lpImageData
		push	ebx

		; --- Output all codes backwards ---
		mov		eax, dwCode
		dec		ebx				;use ebx as counter
		
		.WHILE	TRUE
			mov		eax, [edi+4*eax]
			mov		LastChar, al		
			push	eax
			push	edx
			;--- PixelOffset = (PixelCounter/width)*alignedwidth + (PixelCounter % width) ---
			mov		eax, ebx
			add		eax, [esi].PixelCounter

			cdq
			div		ImageWidth
			mov		ecx, edx
			mul		AlignedWidth
			add		ecx, eax
			mov		edx, ecx
			
			and		cl, 111b	;RealBitOffset  = PixelOffset & 111b;
			mov		al, LastChar
			ror		al, 1
			ror		al, cl		
			
			shr		edx, 3		;RealByteOffset = PixelOffset % 8;
			mov		ecx, edx
			pop		edx
			or		[edx+ecx], al

			pop		eax
			shr		eax, 20
			and		eax, 0fffh
			.BREAK .IF ebx==0
			dec		ebx
		.ENDW
		pop		ebx
		add		[esi].PixelCounter, ebx
		mov		al, LastChar
	
	.ENDIF
ret
DECOutputString	endp
;===============================================================================
; DECIsCodeInTable
;===============================================================================
; Returns TRUE if the given code is in the string table
;
DECIsCodeInTable	proc	uses esi  ebx lpGifInfo:DWORD, dwCode:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	mov		ebx, dwCode
	xor		eax, eax
	mov		ecx, [esi].dwLZWTableCount
	.IF		ecx>ebx
		inc		eax
	.ENDIF
	assume	esi:nothing
ret
DECIsCodeInTable	endp
;===============================================================================
; GetNextCode
;===============================================================================
; retrieves the next code from the compressed data stream and loads new bits
; if needed.
GetNextCode		proc	uses esi edi ebx lpGifInfo:DWORD

	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	;mov		edi, [esi].lpGIFImage
	
	mov		ebx, [esi].CurrentBits

	; --- Create bitmask for current nr of bits ---
	mov		cl,	[esi].dbCurBitSize
	xor		eax, eax
	inc		eax
	shl		eax, cl
	dec		eax
	
	; --- AND ebx with bitmask ---
	and		ebx, eax
	
	
	; --- Shift used bits ---
	mov		cl, [esi].dbCurBitSize
	shr		[esi].CurrentBits, cl
	
	; --- Decrease # of bits available ---
	sub		[esi].dbBitsAvailable, cl
	
	; --- Load as many bits as possible ---
	.WHILE	[esi].dbBitsAvailable<=24
		invoke	LoadNextByte, esi
	.ENDW
	
	; --- Return bits ---
	mov		eax, ebx
	
assume	esi:nothing
ret
GetNextCode		endp

;===============================================================================
; LoadNextByte
;===============================================================================
; Loads the currentbits entry with new bits
;
LoadNextByte	proc	uses esi edi ebx lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
		
	; --- Get next byte from datastream ---
	mov		edi, [esi].lpCurrentByte
	.IF		edi==-1		;end of data already reached
		xor		al, al	;return 0 as padding
		jmp		@gotbyte
	.ENDIF
	;End of subblock reached?
	.IF		edi==[esi].lpLastEOSB 
		xor		eax, eax
		mov		al, [edi]	; new subblock size
		.IF		al==0		; if size=0, end of data
			mov		[esi].lpCurrentByte, -1	;EOF indicator
			xor		al, al
			jmp		@gotbyte
		.ELSE 				; set new end of subblock
			inc		edi		; new subblock with size al
			add		eax, edi
			mov		[esi].lpLastEOSB, eax ;set new end of subblock
			add		[esi].lpCurrentByte, 2
			mov		al, [edi]
		.ENDIF
	.ELSE
		mov		al, [edi]			;get next byte
		inc		[esi].lpCurrentByte	;increase for next read
	.ENDIF

	; --- Got the new byte from the datasteam ---
@gotbyte:
	mov		cl, [esi].dbBitsAvailable	;Get nr of bits available
	mov		edx, [esi].CurrentBits		;Get current bits
	and		eax, 0ffh					;eax = al
	shl		eax, cl						;Shift bits to right position
	or		edx, eax					;OR new bits with current bits
	mov		[esi].CurrentBits, edx		;Save bits
	add		[esi].dbBitsAvailable, 8	;Set new bits avaialble
	assume	esi:nothing
ret
LoadNextByte	endp
;===============================================================================
;	DECInitialize
;===============================================================================
; Initializes the decoder (table for LZW etc)
;
DECInitialize	proc	uses esi edi lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	mov		edi, [esi].lpGIFImage
	; --- Get initial LZW code size ---
	mov		cl, [edi]
	inc 	cl
	mov		[esi].dbInitBitSize, cl
	
	; --- Set current bit size to initial bitsize ---
	mov		[esi].dbCurBitSize, cl
	
	; --- Set size of LZW table ---
	.IF		cl<8		;if initial code size is less than 8-bits, 
		mov 	cl, 8	;set to 8-bits, if higher, don't change it
	.ENDIF
	; Calculate size of codetable:
	xor		eax, eax
	inc		eax
	add		cl, 2		; bitsize + 2
	shl		eax, cl		; eax = 2 ^ (bitsize + 2) = 4 * (2^bitsize)
	mov		[esi].dwLZWTableSize, eax
	
	; --- Allocate memory for LZW Table and store pointer ---
	invoke	GlobalAlloc, GMEM_FIXED, eax
	mov		[esi].lpLZWTable, eax

	
	; --- Reset LZW ---
	invoke	DECResetLZW, esi
	

	
	assume	esi:nothing
xor	eax, eax
inc eax
ret
DECInitialize	endp
;===============================================================================
; DECResetLZW
;===============================================================================
; Resets the LZW Table and variables (this is the proper action for a clear
; code (CC) in data
DECResetLZW		proc	uses esi edi ebx lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:PTR GIF_INFO
	; The LZW table has the following format:
	; An array of DWORDS represents an array of string entries
	; One entry consists of a prefix code (which is an index to
	; another entry that should prefix the entry), a string length,
	; and the suffix byte (this byte should be added to the prefix
	; to get the full string).
	; DWORD:
	; | AAAA AAAA | AAAA BBBB | BBBB BBBB | CCCC CCCC |
	; ^												^
	; MSB                                         LSB
	; AAAAAAAAAAAA:  12-bit index in the LZW Table that indicates the prefix
	; BBBBBBBBBBBB:  12-bit length of the string
	; CCCCCCCC	  :   8-bit suffix byte
	
;int 3
	; --- reset LZW Table ---
	; Reset bit size:
	mov		cl, [esi].dbInitBitSize

	mov		[esi].dbCurBitSize, cl
	; Get initial nr of entries
	;mov		eax, [esi].ColorCount
	;NOTE: THE LINE ABOVE WAS PREVIOUSLY USED TO GET THE NR OF ROOT ENTRIES
	; THIS DIDN'T WORK BECAUSE 1-BIT GIF'S ARE FOR AN UNKNOWN REASON CODED
	; AS 2-BIT GIFS, SO THE NR OF ROOT ENTRIES HAS TO BE DETERMINED FROM THE
	; INITIAL BITSIZE
	dec		cl
	xor		eax, eax
	inc		eax
	shl		eax, cl		; 2**codesize
	add		eax, 2
	mov		[esi].dwLZWTableCount, eax
	mov		edx, [esi].lpLZWTable
	; Fill first {2**codcesize+2} entries
	shl		eax, 2
	add		eax, edx
	xor		ecx, ecx
	inc		ch			;size of sting is set to 1
	.WHILE	edx<eax
		mov		[edx], ecx
		inc		cl
	add	edx, 4
	.ENDW

	; --- set clear code ---
	mov		eax, [esi].dwLZWTableCount
	sub		eax, 2
	mov		[esi].dwLZWClear, eax
	
	mov		[esi].IsFirstCode, 1
	;lpLZWTable		dd	?				; Pointer to LZW Table
	;dwLZWTableSize	dd	?				; Current maximum size of table
	;dwLZWTableCount	dd	?				; Current nr of entries in the LZW table
	;dbCurBitSize	db	?				; Current LZW code size
	;dbInitBitSize	db	?				; Initial LZW code size
	;dwLZWClear		dd	?				; Value of LZW Clear code (CC)
	
	

	
	
	assume	esi:nothing

xor eax, eax
inc	eax
ret
DECResetLZW		endp


;===============================================================================
;	GIFLoadHeader
;===============================================================================
; Loads header and logical screen descriptor
;
GIFLoadHeader	proc uses esi edi ebx lpGifInfo:DWORD
	mov		esi, lpGifInfo
	assume	esi:ptr GIF_INFO
	
	; --- EDI points to GIF Data, EBX is size of data ---
	mov		edi, [esi].lpGIFData
	mov		ebx, [esi].dwGIFDataSize
	
	.IF		ebx<11				; ebx smaller than header & Logical screen descr. -> invalid GIF
		jmp		@failed
	.ENDIF
	
	; --- Check first 3 bytes ("GIF" signature) ---
	mov		eax, [edi]	
	and		eax, 0ffffffh
	.IF		eax!="FIG"			; GIF Magic signature
		jmp		@failed
	.ENDIF
	
	; --- Copy version number ---
	lea		ecx, [esi].GIFVersion
	mov		ax, [edi+3]		
	mov		[ecx], ax
	mov		al, [edi+5]		
	mov		[ecx+2], al
	
	; --- Copy logical screen sizes ---
	xor		eax, eax
	mov		ax, [edi+6]			;Get Logical screen width
	mov		[esi].dwLSWidth, eax
	mov		ax, [edi+8]			;Get Logical screen width
	mov		[esi].dwLSHeight, eax
	
	; --- Set global color table flag and size ---
	mov		cl, [edi+10]		
	mov		al, cl
	rol		al, 1
	and		al, 1
	mov		[esi].fGlobalCTable, al
	and		cl, 111b
	mov		[esi].dbGCTableSize, cl

	assume 	esi:nothing
xor		eax, eax
inc		eax
ret
@failed:
xor		eax, eax
ret
GIFLoadHeader	endp
;===============================================================================
;	GIFInitalizeDecoder
;===============================================================================
; Initializes decoder
;
GIFInitalizeDecoder	proc	uses esi edi ebx lpGifInfo:DWORD
LOCAL	GCTSize:DWORD
	mov		esi, lpGifInfo
	assume	esi:ptr GIF_INFO
	
	; --- EDI points to GIF Data, EBX to end of data ---
	mov		edi, [esi].lpGIFData
	mov		ebx, [esi].dwGIFDataSize
	add		ebx, edi 
	
	add		edi, 13		; skip header and local screen descriptor
	
	; --- Get global color tablesize in bytes ---
	mov		cl, [esi].dbGCTableSize
	inc		cl
	xor		eax, eax	
	inc		eax
	shl		eax, cl
	mov		ecx, eax
	shl		ecx, 1	;--+--> ecx * 3
	add		ecx, eax;--+
	mov		GCTSize, ecx
	
	; --- Skip global color table if available ---
	.IF		[esi].fGlobalCTable!=0
		add		edi, ecx
	.ENDIF
	.IF		edi>=ebx
		jmp		@failed
	.ENDIF

	; --- Search through the GIF blocks for the first graphic block ---
	.WHILE	edi<ebx
		mov		al,	[edi]					; Get block type
		.IF		al==EXTENSIONBLOCK			
			invoke	SkipExtension, edi, ebx
			mov		edi, eax
		.ELSEIF	al==IMAGEDESCRIPTOR
			invoke	ProcessImageDescriptor, lpGifInfo, edi, ebx
			add		edi, 11
			invoke	SkipSubBlocks, edi
			.BREAK
		.ENDIF
	.ENDW
	
	assume 	esi:nothing
xor		eax, eax
inc		eax
ret
@failed:
xor		eax, eax
ret
GIFInitalizeDecoder	endp

;===============================================================================
; ProcessImageDescriptor
;===============================================================================
; Processes an image desriptor block
ProcessImageDescriptor	proc uses esi edi lpGIFInfo:DWORD, lpStart:DWORD, lpEOF:DWORD
	mov		esi, lpGIFInfo
	assume	esi:ptr GIF_INFO
	mov		edx, lpStart
	xor		eax, eax

	; --- copy image width & height ---	
	mov		ax, [edx+5]
	mov		[esi].dwImageWidth, eax
	mov		ax, [edx+7]
	mov		[esi].dwImageHeight, eax		
	
	; --- look for local color table ---
	mov		cl, [edx+9]
	mov		al, cl
	rol		cl, 1
	and		cl, 1
	mov		[esi].fLocalCTable, cl
	and		al, 111b
	mov		[esi].dbLCTableSize, al
	
	; --- Find out which color table to use ---
	.IF		cl==1		;local color table available
		; --- Set color count ---
		mov		cl, [esi].dbLCTableSize
		inc		cl
		xor		eax, eax
		inc		eax
		shl		eax, cl
		mov		[esi].ColorCount, eax
		; --- Set color table pointer ---
		mov		eax, edx
		add		eax, 10	;move to the local color table
		mov		[esi].lpColorTable, eax
	.ELSE				;no local color table, use global color table
		; --- Set color count ---
		mov		cl, [esi].dbGCTableSize
		inc		cl
		xor		eax, eax
		inc		eax
		shl		eax, cl
		mov		[esi].ColorCount, eax
		; --- Set color table pointer ---
		mov		eax, [esi].lpGIFData
		add		eax, 13 ;move to the global color table (at offset 13 in file)
		mov		[esi].lpColorTable, eax
	.ENDIF
	
	; --- Set pointer to real image data (table based image data)
	mov		eax, edx
	add		eax, 10 				;Skip image descriptor
	.IF		[esi].fLocalCTable==1
		mov		ecx, [esi].ColorCount	;-+
		add		eax, ecx				; + add size of colortable
		shl		ecx, 1					; + (colorcount * 3)
		add		eax, ecx				;-+
	.ENDIF
	mov		[esi].lpGIFImage, eax
	; --- Set some pointers to initialize the decompressor
	inc		eax
	mov		[esi].lpLastEOSB, eax
	mov		[esi].lpCurrentByte, eax
	assume	esi:nothing
xor eax, eax
inc eax
ret
ProcessImageDescriptor	endp



;
; Skips an extension block that starts at lpStart. EOF is at lpEnd.
; returns pointer to the end of block + 1 
; or 0 if failed
SkipExtension proc	lpStart:DWORD, lpEnd:DWORD
	
	; --- Get type of extension ---
	mov		edx, lpStart
	mov		al, [edx+1]
	
	; --- Skip each type seperately ---
	.IF		al==GRAPHICCONTROLEXTENSTION
		; Graphic control extension, has a fixed size of 8 bytes
		add		edx, 8
		mov		eax, edx
	.ELSEIF	al==COMMENTEXTENSION
		; Comment extension, consists of subblocks of text
		add		edx, 2	;move to first subblock
		invoke	SkipSubBlocks, edx
		
	.ELSEIF	al==PLAINTEXTEXTENSION
		; Plain text extension, 15 header bytes and subblocks follow
		add		edx, 15
		invoke	SkipSubBlocks, edx
		
	.ELSEIF al==APPLICATIONEXTENSION
		; Application extension, 14 header bytes and subblocks follow
		add		edx, 14
		invoke	SkipSubBlocks, edx
		
	.ENDIF
ret
SkipExtension endp

;===============================================================================
;	SkipSubBlocks
;===============================================================================
; Skips a sequence of subblocks until a block-terminator is found.
; returns a pointer to the end of the block + 1.
; lpStart points to the start of the subblocks
;
SkipSubBlocks	proc lpStart:DWORD
	mov		eax, lpStart
	xor		ecx, ecx
	.WHILE	TRUE
		mov		cl, [eax]	; get size of subblock
		.BREAK .IF cl==0	; size=0 means terminator
		inc		eax
		add		eax, ecx	; add size to pointer
	.ENDW 
	inc 	eax
ret
SkipSubBlocks	endp


end
